#include <vector>
#include <optional>
#include <string>
#include <string_view>
#include <unordered_map>
#include <iostream>
#include <fstream>
#include <limits>

template <template<class,class,class...> class C, typename K, typename V, typename... Args>
V get_or(const C<K,V,Args...>& m, K const& key, const V & defval)
{
	auto it = m.find(key);
	if (it == m.end()) return defval;
	else return it->second;
}

using tape_t = std::vector<unsigned char>;
struct program_context {
	tape_t tape = tape_t(30'000);
	int ptr{};
	const std::reference_wrapper<std::istream> input;
	const std::reference_wrapper<std::ostream> output;
	enum eofAction {zero, max, nothing} onEof;
	
	program_context(
		std::istream& is, std::ostream& os, eofAction eof
	) : input(is), output(os), onEof(eof) {}
};

std::unordered_map<char, std::function<void(program_context&)>>
	actions = {
		{'<', [](program_context& pc) {
			--pc.ptr;
			if (pc.ptr < 0) {
				pc.ptr = 0;
			}
		}},
		{'>', [](program_context& pc) {
			if (pc.ptr == std::numeric_limits<int>::max()) {
				std::abort();
			}
			++pc.ptr;
			if (pc.ptr > pc.tape.size()) {
				pc.tape.push_back(0);
			}
		}},
		{'+', [](program_context& pc) {
			++pc.tape[pc.ptr];
		}},
		{'-', [](program_context& pc) {
			--pc.tape[pc.ptr];
		}},
		{'.', [](program_context& pc) {
			pc.output.get() << static_cast<char>(pc.tape[pc.ptr]);
		}},
		{',', [](program_context& pc) {
			unsigned char& c = pc.tape[pc.ptr];
			pc.input.get().get(reinterpret_cast<char&>(c));
			if (!pc.input.get()) {
				switch (pc.onEof) {
				case program_context::zero:
					c = 0;
					break;
				case program_context::max:
					c = 1;
					break;
				case program_context::nothing:
					;
				}
			}
			return;
		}},
	};

void interpret(std::string_view, program_context);

void subinterpret(std::string_view program, int ip,
	program_context& pc);

int main(int argc, char** argv)
{
	if (argc == 2 && argv[1] == std::string("-h")) {
		std::cout<<
R"(interface: bf <program_file> <output_file> <input_file>
"-" denotes standard in/output
<program_file> and <output_file> default to "-"
<input_file> defaults to <program_file>
<program_file> is read until a '!', if it exists,
    for the program source
if <input_file> is left as default and no '!' is found in
    <program_file>, it is an error.
)";
		return 0;
	}
	
	std::optional <std::ifstream> pr_f;
	std::istream* pfile_p;
	std::optional <std::ofstream> out_f;
	std::ostream* ofile_p;
	std::optional <std::ifstream> in_f;
	std::istream* infile_p;
	
	if ((argc <= 1) || argv[1] == std::string("-")) {
		pfile_p = &std::cin;
	} else {
		pr_f.emplace(argv[1]);
		pfile_p = &*pr_f;
	}
	auto& pfile = *pfile_p;
	
	if ((argc <= 2) || argv[2] == std::string("-")) {
		ofile_p = &std::cout;
	} else {
		out_f.emplace(argv[2]);
		ofile_p = &*out_f;
	}
	auto& outfile = *ofile_p;
	
	std::string program;
	std::getline(pfile, program, '!');
	
	if (argc <= 3) {
		if (!pfile) std::abort();
		infile_p = pfile_p;
	} else if (argv[3] == std::string("-")) {
		infile_p = &std::cin;
	} else {
		in_f.emplace(argv[1]);
		infile_p = &*in_f;
	}
	auto& infile = *infile_p;
	
	program_context::eofAction act = program_context::zero;
	interpret(program, {infile, outfile, act});
}

auto doLoop = [](program_context& pc, std::string_view program) {
	int loopcounter{}, index{};
	std::for_each(program.begin(), program.end(),
		[&](char c){
			if (c == '[') {
				++loopcounter;
			} else if (c == ']') {
				--loopcounter;
				if (!c) {
					subinterpret(program.substr(0, index), 1, pc);
				}
			}
			++index;
		}
	);
};

void interpret(std::string_view program,
	program_context pc)
{
	int ip{};
	while (ip != program.size()) {
		
	}
}

void subinterpret(std::string_view program, int ip,
	program_context& pc)
{
	
}